/**
 * Contentful Delivery API Client. Contains methods which allow access to the
 * different kinds of entities present in Contentful (Entries, Assets, etc).
 */
import { AxiosInstance } from 'contentful-sdk-core';
import { GetGlobalOptions } from './create-global-options';
import { Asset, AssetCollection, AssetFields, AssetKey, AssetQueries, ContentType, ContentTypeCollection, EntriesQueries, LocaleCollection, LocaleCode, Space, SyncCollection, Tag, TagCollection, Entry, EntryCollection } from './types';
import { EntryQueries } from './types/query/query';
import { FieldsType } from './types/query/util';
import { AddChainModifier, ChainModifiers, ChainOptions } from './utils/client-helpers';
type ClientMethodsWithAllLocales<Modifiers extends ChainModifiers> = {
    getEntry<Fields extends FieldsType, Locales extends LocaleCode = LocaleCode>(id: string, query?: Omit<EntryQueries, 'locale'>): Promise<Entry<Fields, Modifiers, Locales>>;
    getEntries<Fields extends FieldsType, Locales extends LocaleCode = LocaleCode>(query?: EntriesQueries): Promise<EntryCollection<Fields, Modifiers, Locales>>;
    parseEntries<Fields extends FieldsType = FieldsType, Locales extends LocaleCode = LocaleCode>(data: EntryCollection<Fields, 'WITH_ALL_LOCALES' | 'WITHOUT_LINK_RESOLUTION', Locales>): EntryCollection<Fields, Modifiers, Locales>;
    getAsset<Locales extends LocaleCode = LocaleCode>(id: string): Promise<Asset<'WITH_ALL_LOCALES', Locales>>;
    getAssets<Locales extends LocaleCode = LocaleCode>(query?: Omit<AssetQueries<AssetFields>, 'locale'>): Promise<AssetCollection<'WITH_ALL_LOCALES', Locales>>;
};
type ClientMethodsWithoutAllLocales<Modifiers extends ChainModifiers> = {
    withAllLocales: Client<AddChainModifier<Modifiers, 'WITH_ALL_LOCALES'>>;
    getEntry<Fields extends FieldsType>(id: string, query?: EntryQueries): Promise<Entry<Fields, Modifiers>>;
    getEntries<Fields extends FieldsType>(query?: EntriesQueries<Fields>): Promise<EntryCollection<Fields, Modifiers>>;
    parseEntries<Fields extends FieldsType = FieldsType>(data: EntryCollection<Fields, 'WITHOUT_LINK_RESOLUTION'>): EntryCollection<Fields, Modifiers>;
    getAsset(id: string, query?: {
        locale?: string;
    }): Promise<Asset<undefined>>;
    getAssets(query?: AssetQueries<AssetFields>): Promise<AssetCollection<undefined>>;
};
export type Client<Modifiers extends ChainModifiers> = BaseClient & ('WITHOUT_LINK_RESOLUTION' extends Modifiers ? {} : 'WITHOUT_UNRESOLVABLE_LINKS' extends Modifiers ? {} : {
    withoutLinkResolution: Client<AddChainModifier<Modifiers, 'WITHOUT_LINK_RESOLUTION'>>;
    withoutUnresolvableLinks: Client<AddChainModifier<Modifiers, 'WITHOUT_UNRESOLVABLE_LINKS'>>;
}) & ('WITH_ALL_LOCALES' extends Modifiers ? ClientMethodsWithAllLocales<Modifiers> : ClientMethodsWithoutAllLocales<Modifiers>);
export type DefaultClient = Client<undefined>;
interface BaseClient {
    version: string;
    /**
     * Gets a Content Type
     * @category API
     * @example
     * ```javascript
     * const contentful = require('contentful')
     *
     * const client = contentful.createClient({
     *   space: '<space_id>',
     *   accessToken: '<content_delivery_api_key>'
     * })
     *
     * const contentType = await client.getContentType('<content_type_id>')
     * console.log(contentType)
     * ```
     */
    getContentType(id: string): Promise<ContentType>;
    /**
     * Gets a collection of Content Types
     * @category API
     * @example
     * ```javascript
     * const contentful = require('contentful')
     *
     * const client = contentful.createClient({
     *   space: '<space_id>',
     *   accessToken: '<content_delivery_api_key>'
     * })
     *
     * const response = await client.getContentTypes()
     * console.log(response.items)
     * ```
     */
    getContentTypes(query?: {
        query?: string;
    }): Promise<ContentTypeCollection>;
    /**
     * Gets the Space which the client is currently configured to use
     * @category API
     * @example
     * ```javascript
     * const contentful = require('contentful')
     *
     * const client = contentful.createClient({
     *   space: '<space_id>',
     *   accessToken: '<content_delivery_api_key>'
     * })
     * // returns the space object with the above <space-id>
     * const space = await client.getSpace()
     * console.log(space)
     * ```
     */
    getSpace(): Promise<Space>;
    /**
     * Gets a collection of Locales
     * @category API
     * @example
     * ```javascript
     * const contentful = require('contentful')
     *
     * const client = contentful.createClient({
     *   space: '<space_id>',
     *   accessToken: '<content_delivery_api_key>'
     * })
     *
     * const response = await client.getLocales()
     * console.log(response.items)
     * ```
     */
    getLocales(): Promise<LocaleCollection>;
    /**
     * Synchronizes either all the content or only new content since last sync
     * See <a href="https://www.contentful.com/developers/docs/concepts/sync/">Synchronization</a> for more information.
     * <strong> Important note: </strong> The the sync api endpoint does not support include or link resolution.
     * However contentful.js is doing link resolution client side if you only make an initial sync.
     * For the delta sync (using nextSyncToken) it is not possible since the sdk wont have access to all the data to make such an operation.
     * @category API
     * @example
     * ```javascript
     * const contentful = require('contentful')
     *
     * const client = contentful.createClient({
     *   space: '<space_id>',
     *   accessToken: '<content_delivery_api_key>'
     * })
     *
     * const response = await client.sync({
     *   initial: true
     * })
     * console.log({
     *   entries: response.entries,
     *   assets: response.assets,
     *   nextSyncToken: response.nextSyncToken
     * })
     * ```
     */
    sync(query: any): Promise<SyncCollection>;
    /**
     * Gets a Tag
     * @category API
     * @example
     * const contentful = require('contentful')
     *
     * const client = contentful.createClient({
     *   space: '<space_id>',
     *   accessToken: '<content_delivery_api_key>'
     * })
     *
     * const tag = await client.getTag('<asset_id>')
     * console.log(tag)
     */
    getTag(id: string): Promise<Tag>;
    /**
     * Gets a collection of Tags
     * @category API
     * @example
     * const contentful = require('contentful')
     *
     * const client = contentful.createClient({
     *   space: '<space_id>',
     *   accessToken: '<content_delivery_api_key>'
     * })
     *
     * const response = await client.getTags()
     * console.log(response.items)
     */
    getTags(query?: any): Promise<TagCollection>;
    /**
     * Creates an asset key for signing asset URLs (Embargoed Assets)
     * @category API
     * @example
     * const contentful = require('contentful')
     *
     * const client = contentful.createClient({
     *   space: '<space_id>',
     *   accessToken: '<content_delivery_api_key>'
     * })
     *
     * const assetKey = await client.getAssetKey(<UNIX timestamp>)
     * console.log(assetKey)
     */
    createAssetKey(expiresAt: number): Promise<AssetKey>;
}
export interface CreateContentfulApiParams {
    http: AxiosInstance;
    getGlobalOptions: GetGlobalOptions;
}
export type ContentfulClientApi = DefaultClient;
export default function createContentfulApi<OptionType extends ChainOptions>({ http, getGlobalOptions }: CreateContentfulApiParams, options?: OptionType): DefaultClient;
export {};
